##
##  bf-ui.pl - OpenPKG build farm user interface (used as CGI)
##  Copyright (c) 2000-2005 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2005 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;
use OpenPKG::Ctx;
use Data::Dumper;
use IO;
use IO::File;
use Getopt::Long;
use OSSP::cfg;
use CGI;
use String::Divert;
use DBI;
use DBD::SQLite;
use MIME::Base64;
use CGI::GuruMeditation;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   catch and hold time
my $time = time();

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $progname = $ctx->cmdprog() . " " . $ctx->cmdname();
my $prefix = $ctx->prefix();

#   options with defaults
my $opt_config     = "$prefix/etc/openpkg-tools/bf.conf";
my $opt_help       = 0;
my $opt_verbose    = 0;

#   configuration
my $opt_verbose = undef;
my $opt_help       = 0;
my $opt_verbose    = 0;

sub verbose () {
    my ($msg) = @_;
    printf STDERR "$progname:DEBUG: %s\n", $msg if($opt_verbose);
}

#   this code reading config is shared between bf-mk.pl, bf-db.pl,
#   bf-ui.pl and src2make.pl. It is a candidate to be moved into a perl
#   module shared between the three as soon as such a module is created.
#
sub subst ($$)
{
    my ($cfg, $txt) = @_;
    while ($txt =~ m|%{([^}]+)\.([^}]+)}|) {
        my ($section, $k) = ($1, $2);
        my $v = $cfg->{$section}->{$k};
        if (defined $v) {
            $txt =~ s|%{$section.$k}|$v|g;
        }
        else {
            die "undefined variable %{$section.$k}";
        }
    }
    return $txt;
}

sub readconfig($)
{
    my ($cfgfile) = @_;
    my $cfg = {};
    my $osspcfg = new OSSP::cfg::simple;

    die "no config file specified" unless (defined $cfgfile);
    my $io = new IO::File "<$cfgfile";
    die "cannot open \"$cfgfile\" for reading ($!)" if(not defined $io);
    my $txt = ''; $txt .= $_ while (<$io>);
    $io->close();

    $osspcfg->parse($txt) || die "error parsing config file";
    my $tree = $osspcfg->unpack(-index => '.*') || die "error unpacking config file";

    foreach my $section (qw(master shared driver slaves)) {
        foreach my $k (keys %{$tree->{$section}->[1]}) {
            my $v = $tree->{$section}->[1]->{$k}->[1];
            $v =~ s|%{([^}\.]+)}|%{$section.\1}|g; #expand shortcuts with section omitted
            $cfg->{$section}->{$k} = $v;
            &verbose("section=$section, k=$k, v=$v");
        }
    }
    foreach my $section (keys %{$cfg}) {
        foreach my $k (keys %{$cfg->{$section}}) {
            $cfg->{$section}->{$k} = &subst($cfg, $cfg->{$section}->{$k});
        }
    }
    return($cfg);
}

sub usage ()
{
    my ($level) = @_;
    print "$progname:$level: Usage: $progname [options]\n" .
          "Available options:\n" .
          " -C|config=s        specifiy config file ($opt_config)\n" .
          " -H|help            print out this usage page\n" .
          " -V|verbose         enable verbose run-time mode\n";
    exit(1) if (defined $level);
    exit(0);
}

#   command line parsing
my $result = GetOptions(
    'C|config=s'     => \$opt_config,
    'H|help'         => \$opt_help,
    'V|verbose'      => \$opt_verbose
) || &usage("ERROR");
&usage() if ($opt_help);

#   handle config
my $cfg = &readconfig($opt_config);
my $shared      = $cfg->{"shared"}->{"prefix"}      || die "mandatory setting shared.prefix undefined";
my $dbfile      = $cfg->{"master"}->{"dbfile"}      || die "mandatory setting master.dbfile undefined";
my $lockretries = $cfg->{"master"}->{"lockretries"} || 250;

#   create objects
my $cgi  = new CGI;
my $myurl = $cgi->url(-relative => 1);
my $dbh = &dbinit($dbfile);

##
##  Generate View Part
##

if    ($cgi->param("page") eq "css") {
    my $out;
    $out = '';
    $out .= &viewhttp('text/html', '+10s');
    $out .= &viewcss();
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "jpg") {
    my $out;
    $out = '';
    $out .= &viewhttp('image/jpg', '+600s');
    $out .= &viewjpg($cgi->param("name"));
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "gif") {
    my $out;
    $out = '';
    $out .= &viewhttp('image/gif', '+600s');
    $out .= &viewgif($cgi->param("name"));
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "processing") {
    my $out;
    $out = '';
    $out .= &viewhttp(undef, undef, $cgi->param("vp_autorefresh"));
    $out .= &viewhtmlhead();
    $out .= &viewmainform();
    $out .= &viewprocessingform();
    $out .= &viewstatusform();
    $out .= &viewmatrixform();
    $out .= "<hr>";
    print STDOUT $out;
    $out = '';
    $out .= &viewprocessing();
    $out .= &viewhtmltail();
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "status") {
    my $out;
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead();
    $out .= &viewmainform();
    $out .= &viewprocessingform();
    $out .= &viewstatusform();
    $out .= &viewmatrixform();
    $out .= "<hr>";
    print STDOUT $out;
    $out = '';
    $out .= &viewstatus();
    $out .= &viewhtmltail();
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "matrix") {
    my $out;
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead();
    $out .= &viewmainform();
    $out .= &viewprocessingform();
    $out .= &viewstatusform();
    $out .= &viewmatrixform();
    $out .= "<hr>";
    print STDOUT $out;
    $out = '';
    &printmatrix();
    $out .= &viewhtmltail();
    print STDOUT $out;
}
elsif ($cgi->param("page") eq "log") {
    my $out;
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead();
    $out .= &viewmainform();
    $out .= &viewprocessingform();
    $out .= &viewstatusform();
    $out .= &viewmatrixform();
    $out .= "<hr>";
    print STDOUT $out;
    $out = '';
    &printlog($cgi->param("file"), $cgi->param("mode"));
    $out .= &viewhtmltail();
    print STDOUT $out;
}
else {
    my $out;
    $cgi->delete_all();
    $out .= &viewhttp();
    $out .= &viewhtmlhead();
    $out .= &viewmainform();
    $out .= &viewprocessingform();
    $out .= &viewstatusform();
    $out .= &viewmatrixform();
    &viewhtmltail();
    print STDOUT $out;
}
#   die gracefully ;-)
$dbh->disconnect();
undef $dbh;
exit(0);

sub viewhttp ($$$)
{
    my ($type, $expires, $refresh) = @_;
    my $header = {};
    $header->{type} =    $type    || 'text/html';
    $header->{expires} = $expires || '+10s';
    if ($refresh) {
        my $self = $cgi->self_url();
        $header->{refresh} = "$refresh; $self";
    }
    return $cgi->header($header);
}

sub viewhtmlhead ($)
{
    my ($html) = @_;
    my $head = "";
    $head .= "<html>\n";
    $head .= "  <head>\n";
    $head .= "    <link rel=\"stylesheet\" type=\"text/css\" href=\"$myurl?page=css\">\n";
    $head .= "  </head>\n";
    $head .= "  <body>\n";
    #$head .= "    <table class=\"bf\">\n";
    $head .= "      <h1>OpenPKG build farm</h1>\n";
    return $head;
}

sub viewhtmltail ()
{
    my $tail = "";
    #$tail .= "    </table>\n";
    $tail .= "  </body>\n";
    $tail .= "</html>\n";
    return $tail;
}

sub viewcss () {
    my $css = '';

    $css .= "/*\n";
    $css .= "**  bfui.cgi - OpenPKG build farm user interface CGI\n";
    $css .= "*/\n";
    $css .= "\n";
    $css .= "HTML {\n";
    $css .= "    background:      #cccccc;\n";
    $css .= "}\n";
    $css .= "BODY {\n";
    $css .= "    background-image: url($myurl?page=jpg&name=bg);\n";
    $css .= "    border:           1px solid #000000;\n";
    $css .= "    padding:          20px 20px 20px 20px;\n";
    $css .= "}\n";
    $css .= "TABLE.bf {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "}\n";
    $css .= "H1 {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       300%;\n";
    $css .= "}\n";
    $css .= "H2 {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       140%;\n";
    $css .= "}\n";
    $css .= ".bf TD.browse {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.browse SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.query {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.query SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.view {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.view SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.result {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.result SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";

    return $css;
}

sub viewjpg () {
    my ($name) = @_;
    $name .= ".jpg";

    my $jpg = {
        "bg.jpg" => <<'EOT'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EOT
    };

    return decode_base64($jpg->{$name});
}

sub viewgif () {
    my ($name) = @_;
    $name .= ".gif";

    my $gif = {
        "icon-dia.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
eHqyLtAxGBW989KMbV+bxDgJADs=
EOT
        "icon-dot.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMR
eLqs8+3JJaur8OCsLZtNmAAAOw==
EOT
        "icon-ok.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
eLpKvMS9KGE8lTand7cVEyoJADs=
EOT
        "icon-x.gif" => <<'EOT'
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
GHHM+g1KFd9qB2J7M50dh3kJADs=
EOT
    };

    return decode_base64($gif->{$name});
}

sub viewmainform()
{
    my $html;
    $html .= $cgi->start_form();
    $html .= $cgi->submit('page','main');
    $html .= $cgi->end_form;
    return $html;
}

sub viewprocessingform()
{
    my $rv;
    my $html;

    $html .= $cgi->start_form();
    $html .= $cgi->submit('page','processing');
    $html .= "&nbsp;autorefresh&nbsp;";
    $html .= $cgi->popup_menu(
        -name     => "vp_autorefresh",
        -value    => [ "0", "3", "10", "60", "300", "900", "3600" ],
        -labels   => { "0" => "never", "3" => "3 secs", "10" => "10 secs", "60" => "1 min", "300" => "5 min", "900" => "15 min", "3600" => "1h" },
        -default  => "0",
    ) . "\n";
    $html .= $cgi->hidden(-name=>'page', -value => "processing");
    $html .= $cgi->end_form;
    return $html;
}

sub viewstatusform()
{
    my $rv;
    my $html;

    my @color;
    push @color, "";
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_build_color " .
                                   "FROM status " .
                                   "ORDER BY st_build_color;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @color, @{$i}->[0] unless (@{$i}->[0] eq "");
    }

    my @inst;
    push @inst, "";
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_inst_host, st_inst_arch, st_inst_os " .
                                   "FROM status " .
                                   "ORDER BY st_inst_host, st_inst_arch, st_inst_os;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @inst, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    my @class;
    push @class, "";
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT cl_package_class " .
                                   "FROM class " .
                                   "ORDER BY cl_package_class;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @class, @{$i}->[0] unless (@{$i}->[0] eq "");
    }

    my @package;
    push @package, "";
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_package_name, st_package_version, st_package_release " .
                                   "FROM status " .
                                   "ORDER BY st_package_name, st_package_version, st_package_release;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @package, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    $html .= $cgi->start_form();
    $html .= $cgi->submit('page','status');
    $html .= "&nbsp;";
    $html .= $cgi->popup_menu(
        -name     => "vs_inst",
        -value    => \@inst,
        -default  => "",
    ) . "\n";
    $html .= $cgi->popup_menu(
        -name     => "vs_package",
        -value    => \@package,
        -default  => "",
    ) . "\n";
    $html .= $cgi->popup_menu(
        -name     => "vs_color",
        -value    => \@color,
        -default  => "",
    ) . "\n";
    $html .= $cgi->popup_menu(
        -name     => "vs_class",
        -value    => \@class,
        -default  => "",
    ) . "\n";
    $html .= $cgi->hidden(-name=>'page', -value => "status");
    $html .= $cgi->end_form;
    return $html;
}

sub viewmatrixform()
{
    my $html;
    $html .= $cgi->start_form();
    $html .= $cgi->submit('page','matrix');
    $html .= $cgi->hidden(-name=>'page', -value => "matrix");
    $html .= $cgi->end_form;
    return $html;
}

sub printlog()
{
    my ($logfile, $mode) = @_;

    my $html;
    if ($logfile =~ m|\.\.| or $logfile =~ m |^/|) {
        die "PANIC:$0: Intruder detected";
    }

    if ( -f "$shared/log/$logfile" and -r "$shared/log/$logfile" and -s "$shared/log/$logfile" ) {
        my $file;
        my $io = new IO::File "<$shared/log/$logfile" or die "unable to read logfile \"$logfile\"";
        $file .= $_ while (<$io>);
        $io->close();
        $file =~ s|&|&amp;|sg;
        $file =~ s|<|&lt;|sg;
        $file =~ s|>|&gt;|sg;
        $file =~ s;^(Executing\((%|--).+?\):.*?)$;<b>$1</b>;mg;
        $file =~ s|^(\+[^\n]*)$|<font color="#333399">$1</font>|mg;
        $file =~ s;^(.*warn.*)$;<font color="#660000"><b>$1</b></font>;mgi;
        $file =~ s;^(.*(error|abort|fail|fatal|bad|illegal|invalid|usage|denied).*)$;<font color="#cc3333"><b>$1</b></font>;mgi;
        $html .= "<pre>\n$file\n</pre>\n";
    }
    else {
        $html .= "Sorry, the logfile \"$logfile\" does not exist\n";
    }
    print STDOUT $html;
}

sub viewprocessing()
{
    my $rv;
    my $html;

    $rv = $dbh->selectall_hashref("SELECT * FROM processing;", "pr_inst_host");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }

    $html .= "<table class=\"bf\">\n";
    $html .= "<tr>";
    $html .= "<td><b>&nbsp;</b></td>";
    $html .= "<td><b>host-arch-os</b></td>";
    $html .= "<td><b>name-version-release</b></td>";
    $html .= "<td><b>building&nbsp;since</b></td>";
    $html .= "<td><b>duration</b></td>";
    $html .= "<tr>\n";
    for my $i (sort keys %{$rv}) {
         my ($name, $version, $release) = (%{$rv}->{$i}->{pr_package_name}, %{$rv}->{$i}->{pr_package_version}, %{$rv}->{$i}->{pr_package_release});
         my ($host, $arch, $os, $tag) = (%{$rv}->{$i}->{pr_inst_host}, %{$rv}->{$i}->{pr_inst_arch}, %{$rv}->{$i}->{pr_inst_os}, %{$rv}->{$i}->{pr_inst_tag});
         my $inst = $host . "-" . $arch . "-" . $os;
         my $package = $name . "-" . $version . "-" . $release;
         my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";

         $html .= "<tr>";

         # color (constant in realtime processing)
         my $l;
         my $r;
         if ($name eq "") {
             $l = "<td>";
             $r = "</td>";
         }
         else {
             $l = "<td><a href=\"$myurl?page=log&mode=tail&file=$logfile\">";
             $r = "</a></td>";
         }
         $html .= "$l<img src=\"$myurl?page=gif&name=icon-dia\" alt=\"dia\" border=\"0\">$r";

         # host-arch-os
         $html .= "<td><a href=\"$myurl?page=status&submit=status&vs_inst=$host-$arch-$os\">$inst</a>" . "&nbsp;</td>";

         # package (name-version-release) or italic "not building"
         if ($name eq "") {
             $html .= "<td><i>not&nbsp;building</i></td>";
         }
         else {
             $html .= "<td><a href=\"$myurl?page=status&submit=status&vs_package=$package\">$package</a></td>";
         }

         # since or "idle"; duration
         my $heartbeat = %{$rv}->{$i}->{pr_heartbeat};
         if ($heartbeat eq "") {
             $html .= "<td>idle</td>";
             $html .= "<td>&nbsp;</td>";
         }
         else {
            my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($heartbeat);
            $html .= sprintf "<td>%04s-%02s-%02s&nbsp;%02s:%02s:%02s</td>", $year+1900, $mon+1, $mday, $hour, $min, $sec;
            $html .= "<td>";
            my $duration = $time - $heartbeat;
            if ($duration < 60) {
                $html .= $duration . "&nbsp;seconds";
            }
            elsif ($duration < 60 * 60) {
                $html .= int($duration/60) . "&nbsp;minutes";
            }
            elsif ($duration < 60 * 60 * 24) {
                $html .= "<b>" . int($duration/60/60) . "&nbsp;hours</b>";
            }
            elsif ($duration < 60 * 60 * 24 * 7) {
                $html .= "<b>" . int($duration/60/60/24) . "&nbsp;days</b>";
            }
            else {
                $html .= "<b>weeks</b>";
            }
            $html .= "</td>";
         }
         $html .= "</tr>\n";
    }
    $html .= "</table>";
    return $html;
}

#   split inst "bsd1-ix86-freebsd5.3-openpkg" into host="bsd1", arch="ix86", os="freebsd5.3" and tag="openpkg"
sub splitinst ($)
{
    my ($inst) = @_;
    my ($host, $arch, $os, $tag);
    if ($inst =~ m|^(.+)-([^-]+)-([^-]+)-([^-]+)$|) {
        ($host, $arch, $os, $tag) = ($1, $2, $3, $4);
    };
    return ($host, $arch, $os, $tag);
}

#   split package "perl-openpkg-5.8.6-20041203" into name="perl-openpkg", version="5.8.6" and release="20041203"
sub splitpackage ($)
{
    my ($package) = @_;
    my ($name, $version, $release);
    if ($package =~ m|^(.+)-([^-]+)-([^-]+)$|) {
        ($name, $version, $release) = ($1, $2, $3);
    };
    return ($name, $version, $release);
}

sub printmatrix()
{
    my $html;
    my ($sth, $sthpackages, $rv);

    my @classes;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT cl_package_class " .
                                   "FROM class " .
                                   "ORDER BY cl_package_class;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }

    my $classpri = {
        "CORE" => 1, 
        "BASE" => 2, 
        "PLUS" => 3, 
        "EVAL" => 4, 
        "JUNK" => 5, 
        "PRIV" => 6
    };

    my @classeswithpri;
    foreach my $i (@{$rv}) {
        my $c = @{$i}->[0];
        my $p = $classpri->{$c};
        $p = "9" unless(defined $p);
        $c = $p . "/" . $c;
        push @classeswithpri, $c;
    }
    foreach my $i (sort @classeswithpri) {
        my $c = $i;
        $c =~ s|^.*/||;
        push @classes, $c;
    }

    my $supportpri = {
        "deprecated" => 1,
        "obsoleted"  => 2,
        "supported"  => 3,
        "tentative"  => 4,
        "forecasted" => 5
    };
    $rv = $dbh->selectall_hashref("SELECT su_inst_host, su_support FROM support;", "su_inst_host");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }

    my $hostsupport = {};
    foreach my $i (keys %{$rv}) {
        my $s = %{$rv}->{$i}->{"su_support"};
        my $p = $supportpri->{$s} || 6;
        $hostsupport->{$i} = $p;
    }

    my @insts;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_inst_host, st_inst_arch, st_inst_os " .
                                   "FROM status " .
                                   "ORDER BY st_inst_host, st_inst_arch, st_inst_os;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }

    sub bysupport {
        my $supporta = $hostsupport->{$a->[0]} || 7;
        my $supportb = $hostsupport->{$b->[0]} || 7;
        if ($supporta == $supportb) { return $a <=> $b }
        return $supporta-$supportb;
    }

    foreach my $i (sort bysupport @{$rv}) {
        push @insts, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    $sthpackages = $dbh->prepare(
        "SELECT DISTINCT st_package_name, st_package_version, st_package_release " .
        "FROM status,class " .
        "WHERE ( " .
        "        st_package_name = cl_package_name " .
        "    and cl_package_class = ? " .
        ") " .
        "ORDER BY st_package_name, st_package_version, st_package_release " .
        ";"
    );
    if (not $sthpackages) {
        die "ERROR:$0: SQLite error prepare";
    }

    $sth = $dbh->prepare(
        "SELECT st_inst_host, st_inst_arch, st_inst_os, st_build_color " .
        "FROM status,class " .
        "WHERE ( " .
        "        st_package_name = ? " .
        "    and st_package_version = ? " .
        "    and st_package_release = ? " .
        "    and st_package_name = cl_package_name " .
        "    and cl_package_class = ? " .
        ");"
    );
    if (not $sth) {
        die "ERROR:$0: SQLite error prepare";
    }

    $html = '';
    foreach my $class (@classes) {
        $html .= "<h2>Class: $class</h2>\n";
        $html .= "<pre>\n";
        print STDOUT $html;
        $html = '';

        my @packages;
        $rv = $sthpackages->execute($class);
        if (not $rv) {
            die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        }
        $rv = $sthpackages->fetchall_arrayref();
        foreach my $i (@{$rv}) {
            push @packages, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
        }

        foreach my $package (@packages) {
            my ($name, $version, $release) = &splitpackage($package);
            $rv = $sth->execute(
                $name, $version, $release,
                $class
            );
            if (not $rv) {
                die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
            }
            $rv = $sth->fetchall_arrayref();
            if ($#{$rv} >= 0 ) {
                my $row = {};
                for my $i (@{$rv}) {
                    $row->{$i->[0] . "-" . $i->[1] . "-" . $i->[2]} = $i->[3];
                }
                my ($goodrow, $lastsupport, $line);
                $goodrow = 1;
                $lastsupport = undef;
                $line = "";
                foreach my $inst (@insts) {
                    my ($host, $arch, $os, $tag) = &splitinst($inst . "-openpkg"); # bsd1-ix86-freebsd5.3-openpkg
                    my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";

                    # support
                    my $support = $hostsupport->{$host};
                    $line .= " " if (defined $lastsupport and $lastsupport != $support);
                    $lastsupport = $support;

                    # color
                    my $color = $row->{$inst};
                    my $l = "<a href=\"$myurl?page=log&mode=tail&file=$logfile\">";
                    my $r = "</a>";
                    if ($color eq "green") {
                        $line .= "$l<img src=\"$myurl?page=gif&name=icon-ok\" alt=\"green\" border=\"0\">$r";
                    }
                    elsif ($color eq "red") {
                        $line .= "$l<img src=\"$myurl?page=gif&name=icon-x\" alt=\"red\" border=\"0\">$r";
                        $goodrow = 0 if ($class eq "CORE" and $support >= 2 and $support <= 4);
                        $goodrow = 0 if ($class eq "BASE" and $support == 3);
                        $goodrow = 0 if ($class eq "PLUS" and $support == 3);
                    }
                    else {
                        $line .= "$l<img src=\"$myurl?page=gif&name=icon-dot\" alt=\"$color\" border=\"0\">$r";
                        $goodrow = 0 if ($class eq "CORE" and $support >= 2 and $support <= 4);
                        $goodrow = 0 if ($class eq "BASE" and $support == 3);
                        $goodrow = 0 if ($class eq "PLUS" and $support == 3);
                    }
                }
                $line .= "&nbsp;<a href=\"$myurl?page=status&submit=status&vs_package=$package\">$package</a>\n";
                $html .= "[<a href=\"$myurl?page=status&submit=status&vs_package=$package\">";
                if ($goodrow) {
                    $html .= "<img src=\"$myurl?page=gif&name=icon-ok\" alt=\"green\" border=\"0\">";
                }
                else {
                    $html .= "<img src=\"$myurl?page=gif&name=icon-x\" alt=\"red\" border=\"0\">";
                }
                $html .= "</a>]&nbsp;";
                $html .= $line;
            }
            print STDOUT $html;
            $html = '';
        }
        $html .= "</pre>\n";
        print STDOUT $html;
        $html = '';
    }
}

sub viewstatus()
{
    my ($host, $arch, $os, $tag);   # bsd1-ix86-freebsd5.3-openpkg
    my ($name, $version, $release); # perl-5.8.6-20050111
    my ($color, $class);            # red, CORE
    my ($select, @args, $sth, $rv);
    my $html;

    my $inst = $cgi->param("vs_inst");
    $inst .= "-openpkg" if ($inst ne "");
    ($host, $arch, $os, $tag) = &splitinst($inst);
    ($name, $version, $release) = &splitpackage($cgi->param("vs_package"));
    $color = $cgi->param("vs_color");
    $class = $cgi->param("vs_class");

    $select = '';
    @args = ();
    $select .=     "SELECT * FROM ";
    $select .=     "    status,class ";
    $select .=     "WHERE ( ";
    my @where = ();
    if ($host) {
        push @where, "st_inst_host = ? ";
        push @args, $host;
    }
    if ($arch) {
        push @where, "st_inst_arch = ? ";
        push @args, $arch;
    }
    if ($os) {
        push @where, "st_inst_os = ? ";
        push @args, $os;
    }
    if ($tag) {
        push @where, "st_inst_tag = ? ";
        push @args, $tag;
    }
    if ($name) {
        push @where, "st_package_name = ? ";
        push @args, $name;
    }
    if ($version) {
        push @where, "st_package_version = ? ";
        push @args, $version;
    }
    if ($release) {
        push @where, "st_package_release = ? ";
        push @args, $release;
    }
    if ($color) {
        push @where, "st_build_color = ? ";
        push @args, $color;
    }
    push @where, "st_package_name = cl_package_name ";
    if ($class) {
        push @where, "cl_package_class = ? ";
        push @args, $class;
    }

    $select .= shift(@where) . " ";
    while (my $q = shift(@where)) {
            $select .= "and " . $q . " ";
    }

    $select .=     "    ) ";
    $select .=     "ORDER by cl_package_class,cl_package_name";
    $select .=     ";";

    if ($#args >= 0) {
        $sth = $dbh->prepare($select);
        if (not $sth) {
            die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        }
        $rv = $sth->execute(@args);
        if (not $rv) {
            die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        }
        $rv = $sth->fetchall_arrayref({});
    }
    else {
        $rv = undef;
    }

    $html = '';
    $html .= "<table class=\"bf\">\n";
    $html .= "<tr>";
    $html .= "<td><b>&nbsp;</b></td>";
    $html .= "<td><b>class</b></td>";
    $html .= "<td><b>host-arch-os</b></td>";
    $html .= "<td><b>name-version-release</b></td>";
    $html .= "<td>&nbsp;</td>";
    $html .= "<td><b>CVS</b></td>";
    $html .= "<td>&nbsp;</td>";
    $html .= "<tr>\n";
    for my $i (@{$rv}) {
         $html .= "<tr>";

         my $class = %{$i}->{cl_package_class};
         my ($host, $arch, $os, $tag) = (%{$i}->{st_inst_host}, %{$i}->{st_inst_arch}, %{$i}->{st_inst_os}, %{$i}->{st_inst_tag});
         my ($name, $version, $release) = (%{$i}->{st_package_name}, %{$i}->{st_package_version}, %{$i}->{st_package_release});
         my $inst = "$host-$arch-$os";
         my $package = "$name-$version-$release";
         my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";

         # color
         my $color = %{$i}->{st_build_color};
         my $l = "<td><a href=\"$myurl?page=log&mode=tail&file=$logfile\">";
         my $r = "</a></td>";
         if ($color eq "green") {
             $html .= "$l<img src=\"$myurl?page=gif&name=icon-ok\" alt=\"green\" border=\"0\">$r";
         }
         elsif ($color eq "red") {
             $html .= "$l<img src=\"$myurl?page=gif&name=icon-x\" alt=\"red\" border=\"0\">$r";
         }
         else {
             $html .= "$l<img src=\"$myurl?page=gif&name=icon-dot\" alt=\"$color\" border=\"0\">$r";
         }

         # class
         $html .= "<td><a href=\"$myurl?page=status&submit=status&vs_class=$class\">$class</a>" . "&nbsp;</td>";

         # inst (host-arch-os)
         $html .= "<td><a href=\"$myurl?page=status&submit=status&vs_inst=$inst\">$inst</a>" . "&nbsp;</td>";

         # package (name-version-release)
         $html .= "<td><a href=\"$myurl?page=status&submit=status&vs_package=$package\">$package</a></td>";

         # cvs
         $html .= "<td><a href=\"http://cvs.openpkg.org/rlog?f=openpkg-src/$name/$name.spec\">timeline</a></td>";
         $html .= "<td><a href=\"http://cvs.openpkg.org/getfile/openpkg-src/$name/$name.spec\">specfile</a></td>";
         $html .= "<td><a href=\"http://cvs.openpkg.org/dir?d=openpkg-src/$name\">browse</a></td>";

         $html .= "</tr>\n";
    }
    $html .= "</table>";
    return $html;
}

#   open database and check whether important tables exist
sub dbinit ($)
{
    my ($dbfile) = @_;
    my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile", "", "", { RaiseError => 0, AutoCommit => 1 });
    if (not defined $dbh) { die "$0: \"cannot open database file \"$dbfile\". SQLite error: ".$dbh->errstr."\n"; }

    my $tables = $dbh->selectall_hashref("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;", "name");

    unless ($tables->{status})     { die "$0: \"status\" table missing in database file \"$dbfile\"\n"; }
    unless ($tables->{processing}) { die "$0: \"processing\" table missing in database file \"$dbfile\"\n"; }
    unless ($tables->{class})      { die "$0: \"class\" table missing in database file \"$dbfile\"\n"; }

    return $dbh;
}

__END__

=pod

=head1 NAME

B<bfui> - OpenPKG build farm user interface

=head1 SYNOPSIS

CGI frontend to build farm database.

=head1 DESCRIPTION

  common options

      [-o|outfile=s]     write output to file
      [-C|config=s]      specifiy config file
      [-H|help]          print out this usage page
      [-V|verbose]       enable verbose run-time mode

Accessible through
http://127.0.0.1/openpkg-cgi/openpkg-tools/bf-ui.cgi
after master initialization has been run, see "openpkg bf-mk".

=cut

