##
##  fetch.pl -- OpenPKG Tool Chain
##  Copyright (c) 2000-2005 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2005 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5;
use strict;
use OpenPKG::Ctx;

#   configure optional debugging
use Data::Dumper;
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $prefix = $ctx->prefix();
my $me = $ctx->cmdprog() . " " . $ctx->cmdname();
my $rpm = $ctx->cmdprog() . " " . "rpm";
my $curl = "$prefix/lib/openpkg/curl";

#   FIXME: hack for "openpkg dev"
delete $ENV{'OPENPKG_TOOLS'};

#   command line parsing
if (@ARGV != 1) {
    print STDERR "openpkg:fetch:ERROR: requires an argument\n";
    print STDERR "openpkg:fetch:USAGE: \"$me <name>.spec\"\n";
    exit(1);
}
my $spec = @ARGV[0];
if (not -f $spec) {
    print STDERR "openpkg:fetch:ERROR: spec file \"$spec\" not found\n";
    exit(1);
}

#   determine package name
my $name = $spec;
$name =~ s|\.[^.]+$||;
$name =~ s|^.+/([^/]+)$|$1|;

#   determine source directories
my $local_srcdir  = `$rpm --define 'name $name' --eval '%_sourcedir'`;
my $local_specdir = `$rpm --define 'name $name' --eval '%_specdir'`;
$local_srcdir  =~ s|\n+$||s;
$local_specdir =~ s|\n+$||s;

#   make sure source and spec directory actually exists
if (not -d $local_srcdir) {
    print(STDERR "openpkg:fetch: creating directory $local_srcdir\n");
    system("mkdir $local_srcdir");
}
if (not -d $local_specdir) {
    print(STDERR "openpkg:fetch: creating directory $local_specdir\n");
    system("mkdir $local_specdir");
}

#   determine some RC parameters
my $RC = {};
my $var;
my $vars = '';
foreach $var (qw(
    _dbpath _rpmdir _srcrpmdir _tmppath
    _target
    l_prefix
    l_fetch_mirror_0 l_fetch_mirror_1 l_fetch_mirror_2 l_fetch_mirror_3 l_fetch_mirror_4
    l_fetch_mirror_5 l_fetch_mirror_6 l_fetch_mirror_7 l_fetch_mirror_8 l_fetch_mirror_9
    l_fetch_backup_0 l_fetch_backup_1 l_fetch_backup_2 l_fetch_backup_3 l_fetch_backup_4
    l_fetch_backup_5 l_fetch_backup_6 l_fetch_backup_7 l_fetch_backup_8 l_fetch_backup_9
    _httpproxy _httpport
    _ftpproxy _ftpport
)) {
    $vars .= "${var}=\%{${var}};";
}
my @assign = split(/;/, `$rpm --eval '$vars'`);
foreach my $assign (@assign) {
    if ($assign =~ m|^(\S+)=(.*)$|s) {
        if ($2 ne '%{'.$1.'}') {
             $RC->{$1} = $2;
        }
    }
}

#   parse spec file
my $DEF = {};
my $SRC = {};
open(SPEC, "<$spec");
my $failed = 0;
while (<SPEC>) {
    s|\n+$||s;
    if (m/^([a-zA-Z_][a-zA-Z0-9_]*):\s*(.+?)\s*$/) {
        $DEF->{lc($1)} = $2;
    }
    if (m/^%define\s+([a-zA-Z_][a-zA-Z0-9_]*)\s+(.+?)\s*$/) {
        $DEF->{lc($1)} = $2;
    }
    if (m/^((Source|Patch)[0-9]+?):\s*(.+?)\s*$/i) {
        my ($remote_srcid, $remote_srcurl) = ($1, $3);

        #   determine expanded remote source location
        1 while ($remote_srcurl =~ s|%{?([a-zA-Z_][a-zA-Z0-9_]*)}?|$DEF->{lc($1)}|sge);
        my $remote_srcdir  = '';
        my $remote_srcfile = $remote_srcurl;
        if ($remote_srcfile =~ m|^(.+)/([^/]+)$|) {
            $remote_srcdir  = $1;
            $remote_srcfile = $2;
        }

        #   display remote source location
        my $file = $remote_srcfile;
        $file = substr($file, 0, 40) if (length($file) > 40);
        printf(STDERR "++ %-9s %-40s ", "$remote_srcid:", $file);

        if (-f "$local_srcdir/$remote_srcfile" or -f "$local_specdir/$remote_srcfile") {
            #   source already on local filesystem
            my $size;
            if (-f "$local_srcdir/$remote_srcfile") {
                $size = (stat("$local_srcdir/$remote_srcfile"))[7];
            }
            else {
                $size = (stat("$local_specdir/$remote_srcfile"))[7];
            }
            if ($size > 1024*1024) {
                $size = sprintf("%.1fMB", $size / (1024*1024));
            }
            elsif ($size > 1024) {
                $size = sprintf("%.0fKB", $size / 1024);
            }
            else {
                $size = sprintf("%d", $size);
            }
            print STDERR "...OK [$size]\n";
        }
        else {
            #   fetch source onto local filesystem
            print STDERR "...MISSING\n";

            #   determine prioritized list of possible source locations
            my @fetch_srcdir = ();
            for (my $i = 0; $i <= 9; $i++) {
                my $url = $RC->{"l_fetch_mirror_$i"};
                next if (not defined($url) or $url eq '' or $url eq '-');
                1 while ($url =~ s|%{?([a-zA-Z_][a-zA-Z0-9_]*)}?|$DEF->{lc($1)}|sge);
                push(@fetch_srcdir, $url);
            }
            push(@fetch_srcdir, $remote_srcdir);
            for (my $i = 0; $i <= 9; $i++) {
                my $url = $RC->{"l_fetch_backup_$i"};
                next if (not defined($url) or $url eq '' or $url eq '-');
                1 while ($url =~ s|%{?([a-zA-Z_][a-zA-Z0-9_]*)}?|$DEF->{lc($1)}|sge);
                push(@fetch_srcdir, $url);
            }

            #   try to fetch from all possible locations
            foreach my $fetch_srcdir (@fetch_srcdir) {
                $fetch_srcdir =~ s|/+$||s;
                print STDERR "openpkg:fetch: fetching from $fetch_srcdir/\n";
                my $err;
                if (($err = &fetch_url("$fetch_srcdir/$remote_srcfile", "$local_srcdir/$remote_srcfile")) ne '') {
                    $err = substr($err, 0, 37)."..." if (length($err) > 40);
                    print STDOUT "openpkg:fetch:ERROR: $err\n";
                    next;
                }
                last;
            }
            if (not -f "$local_srcdir/$remote_srcfile") {
                print STDERR "openpkg:fetch: failed to fetch source file `$remote_srcfile'\n";
                $failed++;
            }
        }
    }
}
close(SPEC);
if ($failed > 0) {
    print STDERR "openpkg:fetch: failed to fetch $failed source files\n";
    exit(1);
}
else {
    exit(0);
}

sub fetch_url {
    my ($src, $dst) = @_;

    #   make sure file URLs have a fully-qualified scheme.
    if ($src =~ m|^/.+|) {
        $src = "file://$src"
    }

    #   make sure only schemes curl(1) supports are used.
    if ($src !~ m;^(file|http|ftp)://.+;) {
        return "invalid URL - only file, http and ftp schemes supported";
    }

    my $rc;
    {
        my $hpxy = $RC->{"_httpproxy"};
        my $hprt = $RC->{"_httpport"};
        my $fpxy = $RC->{"_ftpproxy"};
        my $fprt = $RC->{"_ftpport"};
    
        $hprt = 80 unless $hprt > 0;
        $fprt = 21 unless $fprt > 0;
    
        if (defined $hpxy && defined $hprt) {
            $hpxy .= ":$hprt";
        }
        if (defined $fpxy && defined $fprt) {
            $fpxy .= ":$fprt";
        }
    
        local($ENV{'http_proxy'}) = $hpxy;
        local($ENV{'ftp_proxy'}) = $fpxy;
    
        #   try to fetch the URL
        unlink("$dst.hdr");
        $rc = system($curl .
                     " --location" .
                     " --max-time 1800" .
                     " --connect-timeout 20" .
                     " --dump-header $dst.hdr" .
                     " --output '$dst' '$src'");
    
    }

    #   check whether command failed
    if ($rc != 0) {
        return "cURL error";
    }

    #   check whether remote sites failed
    if (-s "$dst.hdr") {
        open(FP, "<$dst.hdr");
        my $response = <FP>;
        close(FP);
        if ($response =~ m|^HTTP/[\d.]+\s+(\d+)|) {
            if ($1 ne 200) {
                $response =~ s|\n$||;
                unlink($dst);
                return $response;
            }
        }
    }

    #   cleanup
    unlink("$dst.hdr");

    return '';
}
